FRANKAKITCHEN_MAPPING = {
    'kitchen_micro_open-v3' : "open microwave",
    'kitchen_sdoor_open-v3' : "slide cabinet",
    'kitchen_ldoor_open-v3' : "open left door",
    'kitchen_light_on-v3' : "switch on light", 
    'kitchen_knob1_on-v3': "turn on stove",
}